import os
import time
import json

import pytest

from multicontents.multi_versions_file_checkpoints import MultiVersionsFileCheckpoints


class TestMultiVersionsFileCheckpoints(object):
    @pytest.fixture
    def root_dir(self, tmpdir):
        return str(tmpdir.mkdir("test"))

    @pytest.fixture
    def checkpoints(self, root_dir):
        return MultiVersionsFileCheckpoints(
            root_dir=root_dir, checkpoint_dir=".checkpoints"
        )

    def test_create_file_checkpoint(self, checkpoints):
        content = "test content"
        path = "inside/folder/file.txt"
        model = checkpoints.create_file_checkpoint(content, "text", path)

        generated_check_point = os.path.join(
            checkpoints.root_dir,
            ".checkpoints",
            "inside__folder__file___txt",
            model["id"],
        )

        with open(generated_check_point) as fp:
            generated_content = fp.read()
        assert content == generated_content

    def test_create_notebook_checkpoint(self, checkpoints):
        content = {}
        path = "inside/folder/file.ipynb"
        model = checkpoints.create_notebook_checkpoint(content, path)

        generated_check_point = os.path.join(
            checkpoints.root_dir,
            ".checkpoints",
            "inside__folder__file___ipynb",
            model["id"],
        )

        with open(generated_check_point) as fp:
            generated_content = fp.read()
        assert content == json.loads(generated_content)

    def test_list_checkpoints__no_checkpoint(self, checkpoints):
        assert checkpoints.list_checkpoints("not exists") == []

    def test_list_checkpoints_has_file_checkpoints(self, checkpoints):
        v1 = checkpoints.create_file_checkpoint("v1", "text", "file.txt")
        time.sleep(1)
        v2 = checkpoints.create_file_checkpoint("v2", "text", "file.txt")

        saved_checkpoints = checkpoints.list_checkpoints("file.txt")
        assert saved_checkpoints == [
            {"id": v2["id"], "last_modified": v2["last_modified"]},
            {"id": v1["id"], "last_modified": v1["last_modified"]},
        ]

    def test_list_checkpoints_has_nb_checkpoints(self, checkpoints):
        v1 = checkpoints.create_notebook_checkpoint({}, "notebook.ipynb")
        time.sleep(1)
        v2 = checkpoints.create_notebook_checkpoint({}, "notebook.ipynb")

        saved_checkpoints = checkpoints.list_checkpoints("notebook.ipynb")
        assert saved_checkpoints == [
            {"id": v2["id"], "last_modified": v2["last_modified"]},
            {"id": v1["id"], "last_modified": v1["last_modified"]},
        ]



import os
import datetime

from jupyter_core.utils import ensure_dir_exists


try:
    from jupyter_server.services.contents.filecheckpoints import GenericFileCheckpoints
except ImportError:
    from notebook.services.contents.filecheckpoints import GenericFileCheckpoints


class MultiVersionsFileCheckpoints(GenericFileCheckpoints):
    def checkpoint_path(self, checkpoint_id, path):
        """find the path to a checkpoint"""
        checkpoint_dir = self.get_checkpoints_path_for_file(path)
        with self.perm_to_403():
            ensure_dir_exists(checkpoint_dir)
        cp_path = os.path.join(checkpoint_dir, checkpoint_id)
        return cp_path

    def get_checkpoints_path_for_file(self, path, create_missing=False):
        path = path.strip("/")
        checkpoint_dir = os.path.join(
            self.root_dir,
            self.checkpoint_dir,
            path.replace("/", "__").replace(".", "___"),
        )
        if create_missing and not os.path.isdir(checkpoint_dir):
            os.makedirs(checkpoint_dir, exist_ok=True)
        return checkpoint_dir

    def get_checkpoint_dir_and_id(self, path, create_missing=False):
        checkpoint_dir = self.get_checkpoints_path_for_file(
            path, create_missing=create_missing
        )
        saved_time = datetime.datetime.now(datetime.timezone.utc)
        checkpoint_id = str(saved_time.timestamp()).replace(".", "")
        return checkpoint_dir, checkpoint_id

    def create_file_checkpoint(self, content, format, path):
        checkpoint_dir, checkpoint_id = self.get_checkpoint_dir_and_id(
            path, create_missing=True
        )
        checkpoint_file_path = os.path.join(checkpoint_dir, checkpoint_id)
        self.log.debug("creating checkpoint for %s", path)
        with self.perm_to_403():
            self._save_file(checkpoint_file_path, content, format=format)
        return self.checkpoint_model(checkpoint_id, checkpoint_file_path)

    def create_notebook_checkpoint(self, nb, path):
        checkpoint_dir, checkpoint_id = self.get_checkpoint_dir_and_id(
            path, create_missing=True
        )
        checkpoint_file_path = os.path.join(checkpoint_dir, checkpoint_id)
        self.log.debug("creating checkpoint for %s", path)
        with self.perm_to_403():
            self._save_notebook(checkpoint_file_path, nb)
        return self.checkpoint_model(checkpoint_id, checkpoint_file_path)

    def list_checkpoints(self, path):
        checkpoint_dir = self.get_checkpoints_path_for_file(path)
        if not os.path.isdir(checkpoint_dir):
            return []
        checkpoints = [
            self.checkpoint_model(
                file_name,
                os.path.join(self.get_checkpoints_path_for_file(path), file_name),
            )
            for file_name in os.listdir(checkpoint_dir)
        ]
        return sorted(checkpoints, key=lambda c: -c["last_modified"].timestamp())



import os
import re
import datetime
import importlib

from tornado.web import HTTPError
from traitlets import Dict

from multicontents.multi_versions_file_checkpoints import MultiVersionsFileCheckpoints

try:
    from jupyter_server.services.contents.manager import ContentsManager
except ImportError:
    from notebook.services.contents.manager import ContentsManager


DUMMY_CREATED_DATE = datetime.datetime.fromtimestamp(86400)


def build_base_model(
    type_,
    path,
    writable=True,
    last_modified=None,
    created=None,
    content=None,
    format=None,
    mimetype=None,
):
    return {
        "name": path.rstrip("/").rsplit("/", 1)[-1],
        "path": path,
        "writable": writable,
        "last_modified": last_modified or DUMMY_CREATED_DATE,
        "created": created or DUMMY_CREATED_DATE,
        "content": content,
        "mimetype": mimetype,
        "format": format,
        "type": type_,
    }


class WrapperManager(object):
    def __init__(self, proxy_path, manager_class, manager_kwargs):
        self.proxy_path = proxy_path
        if isinstance(manager_class, str):
            module_name, cls_name = manager_class.rsplit(".", 1)
            manager_class = getattr(importlib.import_module(module_name), cls_name)
        self.manager = manager_class(**manager_kwargs)

    def is_parent_directory_of(self, path):
        path = path.strip("/")
        proxy_folders = self.proxy_path.split("/") if self.proxy_path != "" else []
        path_folders = path.lstrip("/").split("/")
        return len(path_folders) >= len(proxy_folders) and "/".join(
            proxy_folders
        ) == "/".join(path_folders[: len(proxy_folders)])

    def is_sub_directory_of(self, path):
        path = path.strip("/")
        proxy_parent = ("/" + self.proxy_path).rsplit("/", 1)[0]
        return (
            proxy_parent.lstrip("/") == path.lstrip("/")
            and path.lstrip("/") != self.proxy_path
        )

    def to_actual_path(self, path):
        path = path.strip("/")
        if re.match(f"^{re.escape(self.proxy_path)}$", path):
            return ""
        return re.sub(f"^{re.escape(self.proxy_path)}/", r"/", path).strip("/")

    def to_proxy_path(self, actual_path):
        actual_path = actual_path.strip("/")
        return os.path.join(self.proxy_path, actual_path).strip("/")

    def get(self, path, *args, **kwargs):
        result = self.manager.get(self.to_actual_path(path), *args, **kwargs)
        if result.get("path", None):
            result["path"] = self.to_proxy_path(result["path"])
        if result.get("content", None) and self.dir_exists(path):
            result["content"] = [
                dict(
                    list(model.items()) + [("path", self.to_proxy_path(model["path"]))]
                )
                for model in result["content"]
            ]
        return result

    def save(self, model, path):
        return self.manager.save(model, self.to_actual_path(path))

    def delete_file(self, path):
        return self.manager.delete_file(self.to_actual_path(path))

    def file_exists(self, path=None):
        return self.manager.file_exists(self.to_actual_path(path))

    def dir_exists(self, path):
        return self.manager.dir_exists(self.to_actual_path(path))

    def is_hidden(self, path):
        return self.manager.is_hidden(self.to_actual_path(path))

    def rename_file(self, old_path, new_path):
        self.manager.rename_file(
            self.to_actual_path(old_path), self.to_actual_path(new_path)
        )


class MultiContentsManager(ContentsManager):

    managers = Dict(help="the path to manager_class settings").tag(config=True)

    def __init__(self, *args, **kwargs):
        super(MultiContentsManager, self).__init__(*args, **kwargs)
        self._managers = [
            WrapperManager(path.lstrip("/"), config["manager_class"], config["kwargs"])
            for path, config in self.managers.items()
        ]
        # path are searched based on the depth, so that we will always match
        # to the closest one first
        self._managers.sort(
            key=lambda manager: (
                manager.proxy_path == "",  # non root matched first
                -len(manager.proxy_path.split("/")),  # match deeper folder first
                -len(
                    manager.proxy_path.rsplit("/", 1)[-1]
                ),  # match longest directory name first
            )
        )

    def get_manager(self, path):
        for manager in self._managers:
            if manager.is_parent_directory_of(path):
                return manager
        raise HTTPError(404, f"Manager not found for path: '{path}'")

    def get(self, path, *args, **kwargs):
        try:
            manager = self.get_manager(path)
            current = manager.get(path, *args, **kwargs)
            is_dir = manager.dir_exists(path)
        except HTTPError as e:
            # if root is not configured, we build a virtual directory to list all
            # the defined path
            if path == "/" or path == "":  # jupyterlab use "/", but classic use ""
                current = build_base_model(
                    type_="directory",
                    path="/",
                    format="json",
                    content=[] if kwargs.get("content", None) else None,
                )
                is_dir = True
            else:
                raise e

        # add virtual directory
        if kwargs.get("content", None) and is_dir:
            extra = [
                build_base_model(
                    type_="directory", path=other_manager.to_proxy_path("")
                )
                for other_manager in self._managers
                if other_manager.is_sub_directory_of(path)
            ]
            current["content"] += extra
        return current

    def rename_file(self, old_path, new_path):
        old_manager = self.get_manager(old_path)
        new_manager = self.get_manager(new_path)

        # avoid renaming on virtual folder
        if (
            old_manager.to_actual_path(old_path) == ""
            or new_manager.to_actual_path(new_path) == ""
        ):
            raise HTTPError(400, reason="You cannot rename the virtual directory")

        if old_manager == new_manager:
            old_manager.rename_file(old_path, new_path)
        else:
            model = old_manager.get(old_path)
            new_manager.save(model, new_path)
            if self.dir_exists(old_path):
                for model in self.get(old_path)["content"]:
                    self.rename_file(
                        os.path.join(old_path, model["name"]),
                        os.path.join(new_path, model["name"]),
                    )
            old_manager.delete_file(old_path)

    def save(self, model, path):
        return self.get_manager(path).save(model, path)

    def delete_file(self, path):
        return self.get_manager(path).delete_file(path)

    def file_exists(self, path=None):
        return self.get_manager(path).file_exists(path)

    def dir_exists(self, path):
        if path == "":
            return True
        return self.get_manager(path).dir_exists(path)

    def is_hidden(self, path):
        if path == "":
            return False
        return self.get_manager(path).is_hidden(path)

    def _checkpoints_class_default(self):
        return MultiVersionsFileCheckpoints



import mock
import pytest
from tornado.web import HTTPError

from multicontents.multicontents_manager import MultiContentsManager
from multicontents.multicontents_manager import WrapperManager


class DummyManager(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __getattr__(self, name):
        self.__dict__[name] = mock.Mock()
        return self.__dict__[name]


class TestWrapperManager(object):
    @pytest.fixture
    def mock_import_module(self):
        with mock.patch("importlib.import_module") as m:
            m.return_value.DummyManager = DummyManager
            yield m

    @pytest.mark.parametrize("manager_cls", ["Dummy.DummyManager", DummyManager])
    def test_init(self, mock_import_module, manager_cls):
        wrapper = WrapperManager("", manager_cls, {"kwarg": "foo"})
        assert wrapper.proxy_path == ""
        assert isinstance(wrapper.manager, DummyManager)
        assert wrapper.manager.kwarg == "foo"

    @pytest.mark.parametrize(
        "proxy_path,path,expected_result",
        [
            ("", "foo", True),
            ("", "/foo", True),
            ("", "/foo/bar", True),
            ("", "foo/bar", True),
            ("foo", "/foo", True),
            ("foo", "/foo/bar", True),
            ("foo", "foo/bar", True),
            ("foo", "/foo/bar/baz", True),
            ("foo", "foo/bar/baz", True),
            ("foo", "bar", False),
        ],
    )
    def test_is_parent_directory_of(self, proxy_path, path, expected_result):
        manager = WrapperManager(proxy_path, DummyManager, {})
        assert manager.is_parent_directory_of(path) == expected_result

    @pytest.mark.parametrize(
        "proxy_path,path,expected_result",
        [
            ("", "foo", False),
            ("", "", False),
            ("foo", "/", True),
            ("foo", "foo", False),
            ("foo", "/foo", False),
            ("foo/bar", "foo/bar", False),
            ("foo/bar", "/foo/bar", False),
            ("foo/bar", "/foo", True),
            ("foo/bar", "foo", True),
        ],
    )
    def test_is_sub_directory_of(self, proxy_path, path, expected_result):
        manager = WrapperManager(proxy_path, DummyManager, {})
        assert manager.is_sub_directory_of(path) == expected_result

    @pytest.mark.parametrize(
        "proxy_path,path,expected_result",
        [
            ("", "/", ""),
            ("", "/foo", "foo"),
            ("foo", "/foo", ""),
            ("foo", "/foo/bar", "bar"),
            ("foo", "/foo/bar/baz", "bar/baz"),
            ("foo/bar", "/foo/bar/", ""),
            ("foo/bar", "/foo/bar/baz", "baz"),
        ],
    )
    def test_to_actual_path(self, proxy_path, path, expected_result):
        manager = WrapperManager(proxy_path, DummyManager, {})
        assert manager.to_actual_path(path) == expected_result

    @pytest.mark.parametrize(
        "proxy_path,path,expected_result",
        [
            ("", "", ""),
            ("", "foo", "foo"),
            ("foo", "", "foo"),
            ("foo", "bar/baz", "foo/bar/baz"),
            ("foo/bar", "baz", "foo/bar/baz"),
        ],
    )
    def test_to_proxy_path(self, proxy_path, path, expected_result):
        manager = WrapperManager(proxy_path, DummyManager, {})
        assert manager.to_proxy_path(path) == expected_result

    def test_get(self):
        mock_get = mock.Mock(return_value={"content": [{"path": "foo/bar"}]})
        mock_dir_exists = mock.Mock(return_value=True)
        manager = WrapperManager(
            "proxy", DummyManager, {"get": mock_get, "dir_exists": mock_dir_exists}
        )
        with mock.patch.object(manager, "to_proxy_path") as mock_to_proxy:
            result = manager.get("path")
        assert result["content"] == [{"path": mock_to_proxy()}]

    @pytest.fixture
    def wrapper_maanger(self):
        manager = WrapperManager("proxy_path", DummyManager, {})
        with mock.patch.object(manager, "to_actual_path"):
            yield manager

    def test_save(self, wrapper_maanger):
        wrapper_maanger.save({}, "path")
        wrapper_maanger.manager.save.assert_called_once_with(
            {}, wrapper_maanger.to_actual_path.return_value
        )
        wrapper_maanger.to_actual_path.assert_called_once_with("path")

    def test_delete_file(self, wrapper_maanger):
        wrapper_maanger.delete_file("path")
        wrapper_maanger.manager.delete_file.assert_called_once_with(
            wrapper_maanger.to_actual_path.return_value
        )
        wrapper_maanger.to_actual_path.assert_called_once_with("path")

    def test_file_exists(self, wrapper_maanger):
        wrapper_maanger.file_exists("path")
        wrapper_maanger.manager.file_exists.assert_called_once_with(
            wrapper_maanger.to_actual_path.return_value
        )
        wrapper_maanger.to_actual_path.assert_called_once_with("path")

    def test_dir_exists(self, wrapper_maanger):
        wrapper_maanger.dir_exists("path")
        wrapper_maanger.manager.dir_exists.assert_called_once_with(
            wrapper_maanger.to_actual_path.return_value
        )
        wrapper_maanger.to_actual_path.assert_called_once_with("path")

    def test_rename_file(self, wrapper_maanger):
        wrapper_maanger.to_actual_path.side_effect = lambda path: f"patched-{path}"
        wrapper_maanger.rename_file("old_path", "new_path")

        wrapper_maanger.manager.rename_file.assert_called_once_with(
            "patched-old_path", "patched-new_path"
        )
        wrapper_maanger.to_actual_path.assert_has_calls(
            [mock.call("old_path"), mock.call("new_path")]
        )


class TestMultiContentsManager(object):
    @pytest.fixture
    def manager_without_root(self):
        return MultiContentsManager(
            managers={
                "foo": {
                    "manager_class": DummyManager,
                    "kwargs": {"kwarg1": "value1", "kwarg2": "value2"},
                },
                "other": {
                    "manager_class": DummyManager,
                    "kwargs": {"kwarg3": "value3"},
                },
            }
        )

    @pytest.fixture
    def manager_with_root(self):
        return MultiContentsManager(
            managers={
                "": {
                    "manager_class": DummyManager,
                    "kwargs": {"kwarg1": "value1", "kwarg2": "value2"},
                },
                "child": {
                    "manager_class": DummyManager,
                    "kwargs": {"kwarg3": "value3"},
                },
            }
        )

    def test_init__with_root(self, manager_with_root):
        assert [m.proxy_path for m in manager_with_root._managers] == ["child", ""]

    def test_init__without_root(self, manager_without_root):
        assert [m.proxy_path for m in manager_without_root._managers] == [
            "other",
            "foo",
        ]

    @pytest.mark.parametrize(
        "path,expected_result", [("", ["abc", "def", "child"]), ("child", ["foo"])]
    )
    def test_get_with_manager_with_root(self, manager_with_root, path, expected_result):
        manager_with_root._managers[1].dir_exists = mock.Mock(return_value=True)
        manager_with_root._managers[1].get = mock.Mock(
            return_value={"content": [{"path": "abc"}, {"path": "def"}]}
        )
        manager_with_root._managers[0].get = mock.Mock(
            return_value={"content": [{"path": "foo"}]}
        )
        result = manager_with_root.get(path, content=True)
        assert [item["path"] for item in result["content"]] == expected_result

    @pytest.mark.parametrize(
        "path,expected_result",
        [("/", ["other", "foo"]), ("foo", ["abc", "def"]), ("other", ["ghi"])],
    )
    def test_get_without_manager_with_root(
        self, manager_without_root, path, expected_result
    ):
        manager_without_root._managers[1].get = mock.Mock(
            return_value={"content": [{"path": "abc"}, {"path": "def"}]}
        )
        manager_without_root._managers[0].get = mock.Mock(
            return_value={"content": [{"path": "ghi"}]}
        )
        result = manager_without_root.get(path, content=True)
        assert [item["path"] for item in result["content"]] == expected_result

    def test_rename_file_same_manager(self, manager_with_root):
        manager_with_root._managers[0].rename_file = mock.Mock()
        manager_with_root.rename_file("child/test1", "child/test2")
        manager_with_root._managers[0].rename_file.assert_called_once_with(
            "child/test1", "child/test2"
        )

    @pytest.mark.parametrize(
        "old_path,new_path", [("/foo", "something/new"), ("something", "/foo")]
    )
    def test_rename_file_virtual_direcotry(
        self, manager_without_root, old_path, new_path
    ):
        with pytest.raises(HTTPError):
            manager_without_root.rename_file(old_path, new_path)

    def test_rename_file_different_manager_file(self, manager_with_root):
        manager_with_root._managers[0].get = mock.Mock()
        manager_with_root._managers[0].delete_file = mock.Mock()
        manager_with_root._managers[1].save = mock.Mock()
        with mock.patch.object(manager_with_root, "dir_exists", return_value=False):
            manager_with_root.rename_file("child/test1", "test2")
        manager_with_root._managers[0].get.assert_called_once_with("child/test1")
        manager_with_root._managers[1].save.assert_called_once_with(
            manager_with_root._managers[0].get.return_value, "test2"
        )
        manager_with_root._managers[0].delete_file.assert_called_once_with(
            "child/test1"
        )

    def get_fake_item(self, path):
        items = {"folder_1": {"file_2": None, "folder_3": {"file_4": None}}}
        obj = items
        for key in path.strip("/").split("/"):
            obj = obj[key]
        return obj

    @pytest.fixture
    def mock_dir_exists(self, manager_with_root):
        def dir_exists(path):
            return self.get_fake_item(path) is not None

        with mock.patch.object(manager_with_root, "dir_exists", side_effect=dir_exists):
            yield

    @pytest.fixture
    def mock_get(self, manager_with_root):
        def get(path):
            obj = self.get_fake_item(path)
            content = [{"name": key} for key in obj] if obj is not None else ""
            return {"name": path.rsplit("/", 1)[-1], "path": path, "content": content}

        with mock.patch.object(manager_with_root, "get", side_effect=get):
            manager_with_root._managers[1].get = mock.Mock(side_effect=get)
            yield

    def test_rename_file_different_manager_dir(
        self, manager_with_root, mock_dir_exists, mock_get
    ):
        manager_with_root._managers[1].delete_file = mock.Mock()
        manager_with_root._managers[0].save = mock.Mock()

        manager_with_root.rename_file("folder_1", "child/test_2")

        old_files = [
            "folder_1",
            "folder_1/file_2",
            "folder_1/folder_3",
            "folder_1/folder_3/file_4",
        ]
        manager_with_root._managers[1].get.assert_has_calls(
            mock.call(f) for f in old_files
        )
        manager_with_root._managers[0].save.assert_has_calls(
            mock.call(mock.ANY, f.replace("folder_1", "child/test_2"))
            for f in old_files
        )
        manager_with_root._managers[1].delete_file.assert_has_calls(
            [mock.call(f) for f in old_files], any_order=True
        )



# -*- coding: utf-8 -*-
import os

from setuptools import find_packages
from setuptools import setup


with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md")) as fp:
    long_description = fp.read()

setup(
    name="multicontents",
    version="0.3.0",
    description="providing contents from multiple sources in jupyter notebook",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/lydian/multicontents",
    author="Lydian Lee",
    author_email="lydianly@gmail.com",
    maintainer="Lydian Lee",
    maintainer_email="lydianly@gmail.com",
    packages=find_packages(),
    install_requires=["notebook", "ipykernel"],
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
)



import subprocess
import socket

import pytest
import requests


pid = None


@pytest.fixture(params=["lab", "notebook"])
def jupyter_process(request):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        port = s.getsockname()[1]
    p = subprocess.Popen(
        [
            "jupyter",
            request.param,
            "-y",
            "--no-browser",
            "--ip=0.0.0.0",
            f"--port={port}",
            "--NotebookApp.token=''",
            "--NotebookApp.password=''",
            "--config=example/jupyter_notebook_config.py",
            "--debug",
        ]
    )

    def clean_up():
        p.kill()

    request.addfinalizer(clean_up)
    yield (p, port)


def test_integration(jupyter_process):
    popen, port = jupyter_process
    if popen.poll() is not None:
        r = requests.get(f"http://127.0.0.1:{port}/api/contents").json()
        assert sorted([folder["name"] for folder in r["content"]]) == sorted(
            ["home", "tmp"]
        )
